/*
 *  DefaultGetOptions.java
 *  Copyright (C) 1999,2000  Joshua T. Rowe
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  To contact the author:
 *  email : jrowe3@home.com
 *  web   : http://jrowe.dhs.org/~jrowe/
 */

package jrowe.getopt;

import java.math.BigDecimal;
import java.math.BigInteger;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Set;

import jrowe.util.TokenIterator;

/**
 * This is the default implementation of a <code>GetOptions</code> object.
 * <p>
 *
 * @author Joshua T. Rowe
 * @version 1.5  JUN 23 2000  Added negation support for long options.
 * @version 1.4  MAR 10 2000  Moved from package org.jrowe.getopt to
 *          jrowe.getopt.
 * @version 1.3  FEB 17 2000  Deprecated setGetOptions.  There seems to be
 *          no reason for the <code>Option</code> to know which
 *          <code>GetOptions</code> object it belongs to.
 * @version 1.2  FEB 09 2000  Started implementing abbreviation support.
 * @version 1.1  JAN 23 2000  Preparation for release 0.1
 * @version 1.0  AUG 27 1999  Initial Version.
 */
public class DefaultGetOptions implements GetOptions {
    /*
     * todo
     *  make sure that if an options has been detected once, the same options
     *  object is not used again.
     *
     *  add variable forceOrder
     *  add variable forceExplicitAssignments
     *
     *  add support for abbreviations of long options
     *
     *  document the GNU_COMPATIBLE and POSIX_COMPATIBLE stuff
     *  make method setCompatible(GNU || POSIX)
     *  make method isCompatible(GNU || POSIX)
     *  make method posixMeHarder() ? :) probably not
     *
     *  make method setHandleExceptionsInternally(boolean)
     *  make method setHandleExceptionsInternally(boolean, Runnable)
     *      runnable cannot be null if boolean is true.  Typically
     *      runnable will just do a System.exit(errorCode). Allow runnable
     *      in case cannot do a system exit or there is other stuff that needs
     *      done
     *
     *  allow multiple strings (not just chars) as assignment tokens
     */

    /**
     * This is the default option prefix.
     *
     * @serial
     */
    public static final String DEFAULT_SHORT_OPTION_PREFIX = "-";
    
    /**
     * This is the default long option prefix.
     *
     * @serial
     */
    public static final String DEFAULT_LONG_OPTION_PREFIX = 
            DEFAULT_SHORT_OPTION_PREFIX + DEFAULT_SHORT_OPTION_PREFIX;
    
    /**
     * This is the default end of options token.
     *
     * @serial
     */
    public static final String DEFAULT_END_OF_OPTIONS_TOKEN = 
            DEFAULT_SHORT_OPTION_PREFIX + DEFAULT_SHORT_OPTION_PREFIX;
    
    /**
     * This is string that signifies an assignment of a value to an option.
     * <p>
     *
     * @serial
     */
    public static final String DEFAULT_ASSIGNMENT_STRING = "=";

    /**
     * This is string that signifies standard input or standard output.
     * <p>
     *
     * @serial
     * @deprecated
     */
    public static final String DEFAULT_STDIO_TOKEN = "-";

    /**
     * This is string that delimits items in a list.
     * <p>
     *
     * @serial
     */
    public static final String DEFAULT_LIST_DELIMITER = ",";

    /**
     * the Default Negation Prefix property.
     *
     * @serial
     *
     * @see #gegationPrefix
     * @see #setNegationPrefix
     * @see #getNegationPrefix
     */
    public static final String DEFAULT_NEGATION_PREFIX = "no";
    
    /**
     * Stores the options that are parseable.
     *
     * @serial
     */
    private Collection optionList;

    /**
     * Stores the number of options that are parseable.
     *
     * @serial
     */
//FIXME what is this for and why do I need it?
    private int optionListSize;

    /**
     * Stores the options that have been parsed.
     *
     * @serial
     */
    private Collection parsedOptionList;

    /**
     * A <code>Map</code> used to store the <code>Option</code> objects.
     * <p>
     *
     * The <code>Option</code> object is the value stored, and it is keyed
     * on by both the option's long name and the option's short name.
     * <p>
     *
     * @serial
     */
    private Map optionsMap;

    /**
     * A <code>StringBuffer</code> that is used to reparse all of the
     * command line arguments.
     *
     * @serial
     */
    private StringBuffer buffer;

    /**
     * The list of tokens.
     * <p>
     *
     * This is analagous to the <code>args</code> parameter in the
     * <code>main</code> method.
     *
     * @serial
     *
     * @see #setTokenList(List)
     * @see #getTokenList()
     */
    private List tokenList;

    /**
     * An <code>Iterator</code> over the tokens passed in via the
     * <code>setArgs</code> method.
     *
     * @serial
     */
    private ListIterator tokenListIterator;

    /**
     * This is the short option prefix.
     *
     * @serial
     *
     * @see #DEFAULT_SHORT_OPTION_PREFIX
     * @see #DEFAULT_LONG_OPTION_PREFIX
     */
    private String shortOptionPrefix;

    /**
     * This is the long option prefix.
     * <p>
     *
     * By default, the long option prefix is the short option prefix
     * doubled.
     *
     * @serial
     *
     * @see #DEFAULT_SHORT_OPTION_PREFIX
     * @see #DEFAULT_LONG_OPTION_PREFIX
     */
    private String longOptionPrefix;

    /**
     * This is the end of options token.
     * <p>
     *
     * By default, the end of options tokens is the short option prefix
     * doubled.
     *
     * @serial
     *
     * @see #DEFAULT_END_OF_OPTIONS_TOKEN
     */
    private String endOfOptionsToken;

    /**
     * This is string that signifies an assignment of a value to an option.
     * <p>
     *
     * @serial
     *
     * @see #DEFAULT_ASSIGNMENT_STRING
     */
    private String assignmentString;

    /**
     * This is string that signifies standard input.
     * <p>
     *
     * @serial
     *
     * @see #DEFAULT_STDIO_TOKEN
     * @deprecated
     */
    private String stdioString;

    /**
     * This is string that delimits items in a list.
     * <p>
     *
     * @serial
     *
     * @see #DEFAULT_LIST_DELIMITER
     */
    private String listDelimiter;

    /**
     * The list of program arguments.
     *
     * @serial
     */
    private List programArgs;

    //DEBUG & DEVEL flags
    //private boolean doubleForLong = true;

    /**
     * Indicates whether long options will be considered valid or not.
     *
     * @serial
     *
     * @see #setLongOptionsEnabled(boolean)
     * @see #getLongOptionsEnabled()
     */
    private boolean longOptionsEnabled = true;

    /**
     * Indicates whether explicit assignments are allowed.
     *
     * @serial
     *
     * @see #setAllowExplicitAssignments(boolean)
     * @see #getAllowExplicitAssignments()
     */
    private boolean allowExplicitAssignments = true;

    /**
     * The default value for the Prefer Long Options property.
     *
     * <p>
     * This value is <code>true</code>.
     * </p>
     *
     * @serial
     */
    private static boolean DEFAULT_PREFER_LONG_OPTIONS = true;

    /**
     * Indicates whether long options will be checked first during parsing.
     *
     * @serial
     *
     * @see #DEFAULT_PREFER_LONG_OPTIONS
     * @see #setPreferLongOptions(boolean)
     * @see #isPreferLongOptions()
     */
    private boolean preferLongOptions;
    
    /**
     * the Negation Prefix property.
     *
     * @serial
     *
     * @see #DEFAULT_NEGATION_PREFIX
     * @see #setNegationPrefix
     * @see #getNegationPrefix
     */
    private String negationPrefix;
    
    /**
     * Creates a <code>DefaultGetOptions</code> object with the default
     * attributes.
     */
    public DefaultGetOptions() {
        super();

        init();

        optionsMap = new HashMap();
        programArgs = createList();
        optionList = createList();
        parsedOptionList = createList();
        setTokenList(createList());
    }
    
    private void init() {
        setShortOptionPrefix(DEFAULT_SHORT_OPTION_PREFIX);
        setLongOptionPrefix(DEFAULT_LONG_OPTION_PREFIX);
        setEndOfOptionsToken(DEFAULT_END_OF_OPTIONS_TOKEN);
//        setAssignmentString(DEFAULT_ASSIGNMENT_STRING);
        assignmentString = DEFAULT_ASSIGNMENT_STRING;
        stdioString = DEFAULT_STDIO_TOKEN;
        listDelimiter = DEFAULT_LIST_DELIMITER;
        setNegationPrefix(DEFAULT_NEGATION_PREFIX);
        setPreferLongOptions(DEFAULT_PREFER_LONG_OPTIONS);
    }

    private List createList() {
        return new LinkedList();
    }

    /**
     * Sets the Token List property.
     *
     * @param newtokenList the new value for the Token List property
     *
     * @see #getTokenList()
     */
    public void setTokenList(List newtokenList) {
        tokenList = newtokenList;
    }

    /**
     * Returns the Token List property.
     *
     * @returns the Token List property
     *
     * @see #setTokenList(List)
     */
    public List getTokenList() {
        return tokenList;
    }

    /**
     * Sets the Prefer Long Options property.
     *
     * @param newpreferLongOptions the new value for the Prefer Long Options property
     *
     * @see #isPreferLongOptions()
     */
    public void setPreferLongOptions(boolean newpreferLongOptions) {
        preferLongOptions = newpreferLongOptions;
    }

    /**
     * Returns the Prefer Long Options property.
     *
     * @returns the Prefer Long Options property
     *
     * @see #setPreferLongOptions(boolean)
     */
    public boolean isPreferLongOptions() {
        return preferLongOptions;
    }

    /**
     * Sets the Long Options Enabled property.
     *
     * @param longOptionsEnabled the new value for the Long Options Enabled property
     */
    public void setLongOptionsEnabled(boolean longOptionsEnabled) {
        this.longOptionsEnabled = longOptionsEnabled;
    }

    /**
     * Returns the Long Options Enabled property.
     *
     * @returns the Long Options Enabled property
     */
    public boolean isLongOptionsEnabled() {
        return this.longOptionsEnabled;
    }

    /**
     * Sets the Short Option Prefix property.
     *
     * @param newShortOptionPrefix the new value for the Short Option Prefix property
     */
    public void setShortOptionPrefix(String newShortOptionPrefix) {
        shortOptionPrefix = newShortOptionPrefix;
    }

    /**
     * Returns the Short Option Prefix property.
     *
     * @returns the Short Option Prefix property
     */
    public String getShortOptionPrefix() {
        return shortOptionPrefix;
    }

    /**
     * Sets the Long Option Prefix property.
     *
     * @param newLongOptionPrefix the new value for the Long Option Prefix property
     */
    public void setLongOptionPrefix(String newLongOptionPrefix) {
        longOptionPrefix = newLongOptionPrefix;
    }

    /**
     * Returns the Long Option Prefix property.
     *
     * @returns the Long Option Prefix property
     */
    public String getLongOptionPrefix() {
        return longOptionPrefix;
    }

    /**
     * Sets the End Of Options Token property.
     *
     * @param newEndOfOptionsToken the new value for the End Of Options Token property
     */
    public void setEndOfOptionsToken(String newEndOfOptionsToken) {
        endOfOptionsToken = newEndOfOptionsToken;
    }

    /**
     * Returns the End Of Options Token property.
     *
     * @returns the End Of Options Token property
     */
    public String getEndOfOptionsToken() {
        return endOfOptionsToken;
    }

    /**
     * Sets the Std IOString property.
     *
     * @param newStdIOString the new value for the Std IOString property
     * @deprecated
     */
    public void setStdIOString(String newStdIOString) {
        stdioString = newStdIOString;
    }

    /**
     * Returns the Std IOString property.
     *
     * @returns the Std IOString property
     * @deprecated
     */
    public String getStdIOString() {
        return stdioString;
    }

    /**
     * Sets the Allow Explicit Assignments property.
     *
     * @param newAllowExplicitAssignments the new value for the Allow Explicit Assignments property
     */
    public void setAllowExplicitAssignments(boolean newAllowExplicitAssignments) {
        allowExplicitAssignments = newAllowExplicitAssignments;
    }

    /**
     * Returns the Allow Explicit Assignments property.
     *
     * @returns the Allow Explicit Assignments property
     */
    public boolean isAllowExplicitAssignments() {
        return allowExplicitAssignments;
    }

    /**
     * Sets the Negation Prefix property.
     *
     * @param newNegationPrefix the new value for the Negation Prefix property
     */
    public void setNegationPrefix(String newNegationPrefix) {
        negationPrefix = newNegationPrefix;
    }

    /**
     * Returns the Negation Prefix property.
     *
     * @returns the Negation Prefix property
     */
    public String getNegationPrefix() {
        return negationPrefix;
    }

    /**
     * An <code>Iterator</code> over the tokens passed in via the
     * <code>setArgs</code> method.
     *
     * @return an <code>Iterator</code> object.
     */
    protected ListIterator getTokenListIterator() {
        return tokenListIterator;
    }

    /**
     * Adds an <code>Option</code> to the list of parseable options.
     *
     * @param option The <code>Option</code> to add.
     */
    public void add(Option option) {
        optionList.add(option);

        OptionRecord or = new OptionRecord();
        or.setOption(option);
        or.setRequestedIndex(optionListSize);

        if ( longOptionsEnabled ) {
            optionsMap.put(option.getName(), or);

//FIXME make variable allowNegation property

            /*
             * if the arg type is boolean then put an alias
             * to the argument with the negationPrefix into the map
             */
            if ( option.getArgType() == Option.BOOLEAN_ARG ) {
                optionsMap.put(getNegationPrefix() + option.getName(), or);
            }

        }
        optionsMap.put(option.getCharacter().toString(), or);

        optionListSize++;
    }

    /**
     * Removes an <code>Option</code> from the list of parseable options.
     *
     * @param optionName The name of the <code>Option</code> to remove.
     * @param The <code>Option</code> which was removed.  If this object does
     *        not contain an </code>Option</code> by that name, then 
     *        <code>null</code> will be returned.
     */
    public Option remove(String optionName) {
        Option option = get(optionName);
        remove(option);
        return option;
    }

    /**
     * Removes an <code>Option</code> from the list of parseable options.
     *
     * @param option The <code>Option</code> to remove.
     */
    public void remove(Option option) {
        optionList.remove(option);
        optionListSize--;

        optionsMap.remove(option.getName());

        optionsMap.remove(option.getCharacter().toString());

        /*
         * if the arg type is boolean then put an alias
         * to the argument with the negationPrefix into the map
         */
        if ( option.getArgType() == Option.BOOLEAN_ARG ) {
            optionsMap.remove(getNegationPrefix() + option.getName());
        }
    }

    /**
     * Returns an <code>Option</code> from the list of parseable options.
     *
     * @param optionName The name of the <code>Option</code> to return.
     * @param The <code>Option</code>.  If this object does not contain an
     *        </code>Option</code> by that name, then <code>null</code> will be
     *        returned.
     */
    public Option get(String optionName) {
boolean print = false && "noaBoolean".equals(optionName);

        Option option = null;

        String key = ( optionName.startsWith(getNegationPrefix()) )
                     ? optionName.substring(getNegationPrefix().length())
                     : optionName;

boolean alg0 = true;
if ( ! alg0 ) {
//FIXME make smarter. what if option name is "none"
System.out.println();

//fixing the no problem
//        OptionRecord or = (OptionRecord) optionsMap.get(key);
        OptionRecord or = getOptionRecord(key);
        Option opt = or.getOption();

System.out.println("key=" + key + "; option=" + opt);
} else {

if ( print ) { System.out.println("optionName=" + optionName); }

        Option opt;
        Iterator it = optionList.iterator();
        while ( it.hasNext() && (option == null)) {
            opt = (Option) it.next();

if ( print ) { System.out.println("option=" + opt.getName()); }

//            if ( optionName.equals(opt.getName()) ) {
            if ( key.equals(opt.getName()) ) {
                option = opt;
if ( print ) { 
    System.out.println("FOUND:option=" + opt);
    new Exception("!!").printStackTrace();
}

            }
        }

}

if ( "version".equals(optionName) ) {
    System.out.println("get(String):option=" + option);
}

        return option;
    }

//uses list via iterator
    /**
     * Returns an <code>Option</code> from the list of parseable options.
     *
     * @param optionChar The char property of the <code>Option</code> to return.
     * @param The <code>Option</code> specified.
     *        If this object does not contain an </code>Option</code> with the
     *        specified property, then <code>null</code> will be returned.
     */
    public Option get(char optionChar) {
        Option option = null;
        Iterator it = iterator();
        Option opt;
        while ( it.hasNext() && (option == null)) {
            opt = (Option) it.next();
            if ( optionChar == opt.getChar() ) {
                option = opt;
            }
        }
        return option;
    }

//uses list
    /**
     * Returns an <code>Iterator</code> object which will iterate over all
     * of the <code>Options</code> contained by this object.
     *
     * @return An <code>Iterator</code>
     */
    public Iterator iterator() {
        return optionList.iterator();
    }

    /**
     * Sets the arguments from the command line.
     * <p>
     *
     * The options are processed immediately.
     *
     * @param args The arguments from the command line.
     *
     * @exception OptionException if there is an error while parsing the
     *            arguments
     */
    public void setArgs(String[] args) throws OptionException {
//FIXME get WHITESPACE_CHAR from the tokenizer
        final char WHITESPACE_CHAR = ' ';

        setTokenList(createList());
        buffer = new StringBuffer();

        for ( int i = 0; i < args.length; i++ ) {
            buffer.append(args[i]);

            if ( ( i + 1 ) < args.length ) {
                buffer.append(WHITESPACE_CHAR);
            }
        }

        /* populate tokenList */
//FIXME use WHITESPACE_CHAR in StringTokenizer
//FIXME don't use string tokenizer to parse string doesn't do quotes right

        StringTokenizer stok = new StringTokenizer(buffer.toString());
        while ( stok.hasMoreTokens() ) {
            getTokenList().add(stok.nextToken());
        }

        parseArgs();
    }

    /**
     * Handles an <code>OptionException</code>.
     *
     * @param e the exception
     */
    protected void handleOptionException(OptionException e) {
        System.err.println(e.getMessage());
    }
   
    /**
     * Returns the program arguments.
     *
     * @return The program arguments.
     */
    public List getProgramArgs() {
        return Collections.unmodifiableList(programArgs);
    }
    
    /**
     * Parses the arguments specified by the <code>setArgs</code> method.
     */
    protected void parseArgs() throws OptionException {
        tokenListIterator = getTokenList().listIterator();
        String token;
        /*
         * In a nested option, this iterator will need to have been handed in.
         */
        Iterator it = getTokenListIterator();
        while ( it.hasNext() ) {
            token = it.next().toString();

            if ( isEndOfOptionsToken(token) ) {
                while ( it.hasNext() ) {
                    programArgs.add(it.next());
                }

            } else if ( isOption(token) ) {
                getOption(token);

            } else {
                programArgs.add(token);
            }

        }

        /* call the run methods here */
        it = getParsedOptions().iterator();
        while ( it.hasNext() ) {
            ((Option) it.next()).runMethod();
        }

    }

    /**
     * Returns whether the specified string is an option.
     *
     * @param s The specified string.
     * @return <code>true</code> if the specified string is an option;
     *         <code>false</code> if the specified string is either a
     *         program argument or an option argument.
     */
    protected boolean isOption(String s) {
        return ( isShortOption(s) || isLongOption(s) );
    }

    protected boolean isShortOption(String s) {
        return ( s.length() > shortOptionPrefix.length() ) &&
               ( s.startsWith(shortOptionPrefix) );
    }

    protected boolean isLongOption(String s) {
        return ( longOptionsEnabled ) &&
               ( s.length() > longOptionPrefix.length() ) &&
               ( s.startsWith(longOptionPrefix) );
    }

    /**
     * Returns whether the specified string indicates that only program
     * arguments follow it.
     *
     * @param s The specified string.
     * @return <code>true</code> if the specified string indicates that only
     *         program arguments follow it.
     */
    protected boolean isEndOfOptionsToken(String s) {
        return endOfOptionsToken.equals(s);
    }
 
    /**
     * Returns the correct <code>OptionRecord</code> given an option id.
     *
     * @param optionID this is either the name or the character property
     *        on an <code>Option</code> object
     * @return the correct <code>OptionRecord</code>.  If no matching
     *         <code>OptionRecord</code> exists, then <code>null</code> is
     *         returned.
     */
    protected OptionRecord getOptionRecord(String optionID) {
        OptionRecord or = (OptionRecord) optionsMap.get(optionID);

        boolean returnValue = ( or != null ) && ( ! or.isParsed() );

        return ( returnValue ) ? or : null;
    }

    /**
     * Returns the appropriate <code>Option</code> object(s) for the given
     * token.
     *
     * @param token a token as if from the <code>setArgs</code> method.
     * @return a <code>List</code> of the matching <code>Option</code> objects.
     * @throws an <code>OptionException</code> if the given option
     *         doesn't exist
     */
    protected List getOption(String token) throws OptionException {
        /*
         * if there are missing options, then throw exception
         */

        List options = null;

        String strippedToken = stripOption(token);

        OptionRecord or = getOptionRecord(strippedToken);

        Object arg = null;

        /*
         * check for concatenated short options
         */
        if ( or == null ) {
            options = getShortOptions(token);

        } else {
            options = createList();
            options.add(or.getOption());
            or.incrementParsedCount();
//boolean problem
//            getArg(or.getOption(), or.getOption().getName(), strippedToken);
            getArg(or.getOption(), strippedToken, strippedToken);

//FIXME moved to getArg method
//                or.getOption().setArg(arg2);
        }


        return options;
    }

    /**
     * Parses a token consisting of concatenated short options and possibly
     * an argument.
     *
     * @param token the token
     * @return the <code>List</code> of <code>Option</code> objects
     */
    private List getShortOptions(String token) throws OptionException {
        /*
         * for posixly correct throw error is find a short option
         *      that takes an arg but isn't at the end of the
         *      concatenation ??
         */

        String strippedToken = stripOption(token);

        List options = createList();

        OptionRecord or;
        String c;
        for ( int i = 0; i < strippedToken.length(); i++ ) {
            c = new String(new char[] { strippedToken.charAt(i) });
            or = (OptionRecord) optionsMap.get(c);

            if ( or == null ) {
                throw new OptionException(
                        "Unknown option '" + c + "'", 
                        OptionException.UNKNOWN_OPTION);
            }
            options.add(or.getOption());
//FIXME increment parsed count don't just assume max instances == 1
//            or.setParsed(true);
            or.incrementParsedCount();

            /*
             *  substring the stripped token so that the getArg method
             *  only deals with the relevant part.
             */
            Object arg = getArg(or.getOption(), 
                    or.getOption().getCharacter().toString(), 
                    strippedToken.substring(i));
//FIXME moved to getArg method
        /*
         *  This is the only place where I need getArg to return a value
         *  Investigate further.
         */
//                or.getOption().setArg(arg);

            /* if arg was part of this token then stop processing this token */
            if ( ( arg != null ) && 
                 ( strippedToken.endsWith(arg.toString()) ) ) {

                break;
            }

        }

        return options;
    }

    /**
     * Turns an arg from a <code>String</code> into the appropriate
     * actual type.
     */
    private Object processArgs(Option opt, String arg) {
        Object processedArg = null;

        switch ( opt.getArgType() ) {
        case Option.SCALAR_ARG :
            /* falls through */

        case Option.STRING_ARG :
            processedArg = arg.toString();
            break;

//FIXME maybe rework so as to not repeat code for numbers
        case Option.INTEGER_ARG:
            try {
                processedArg = new BigInteger(arg);

                try {
                    processedArg = new Long(Long.parseLong(arg));

                } catch (NumberFormatException e) {
                    /* if number is too big for a long */
                }

            } catch (NumberFormatException e) {
                throw new OptionException(
                        "arg '" + arg + "' is not a valid integer",
                        OptionException.MALFORMED_ARGUMENT);
            }

            break;
            
        case Option.REAL_ARG:
            /* falls through */

        case Option.NUMERIC_ARG :
            try {
                BigDecimal bigD = new BigDecimal(arg);
                Double littleD = null;

                BigDecimal bigD2 = null;
                try {
                    littleD = Double.valueOf(arg);
                    bigD2 = new BigDecimal(littleD.doubleValue());

                } catch (NumberFormatException e) {
                }

                processedArg = ( ( bigD2 != null ) && ( bigD.equals(bigD2) ) )
                               ? (Object) littleD
                               : (Object) bigD;

            } catch (NumberFormatException e) {
                throw new OptionException(
                        "arg '" + arg + "' is not a valid number",
                        OptionException.MALFORMED_ARGUMENT);
            }

            break;

        case Option.LIST_ARG :
            StringTokenizer stok = new StringTokenizer(arg, listDelimiter);
            List l = createList();
            while ( stok.hasMoreTokens() ) {
                l.add(stok.nextToken());
            }
            processedArg = l;
            break;

        case Option.BOOLEAN_ARG :
            if ( arg.equalsIgnoreCase(Boolean.TRUE.toString()) ||
                 arg.equalsIgnoreCase(Boolean.FALSE.toString()) ) {
                processedArg = new Boolean(arg);

            } else {
                throw new OptionException(
                        "arg '" + arg + "' is not a valid boolean",
                        OptionException.MALFORMED_ARGUMENT);
            }

            break;

        case Option.COUNT_ARG :
            Number n = (Number) opt.getArg();
            long val = n.longValue();
            val++;
            opt.setArg(new Long(val));
            break;

        case Option.NO_ARG :
            /* fall through */

        default :
            processedArg = null;
            break;
        }

        return processedArg;
    }

    /**
     * Parses and returns the argument that ought to be assigned to the
     * given option.
     *
     * @param opt the given <code>Option</code>
     * @param token the given token
     * @return the argument to be assigned to the <code>Option</code>
     *
     * @exception IllegalArgumentException if any of the parameters are null
     */
//FIXME rename this
    protected Object getArg(Option opt, String optID, String token) {
        if ( opt == null ) {
            throw new IllegalArgumentException("opt cannot be null");
        }

        if ( optID == null ) {
            throw new IllegalArgumentException("optID cannot be null");
        }

        if ( token == null ) {
            throw new IllegalArgumentException("token cannot be null");
        }

        /*
         * Must distinguish between a short option and a long option
         *      probably should move this functionality into the respective
         *      methods.
         */

        Object arg = null;

        /*
         * figure out whether the option can take an argument
         * note : if the option has an optional argument, 
         *        hasArg should be true
         */
        boolean hasArg = DefaultOption.hasArg(opt.getArgType());
        if ( hasArg ) {

            /*
             * does the token might have the argument embedded in it
             * ie : token = 'optionFoo=argumentBar'
             */
            if ( token.length() > optID.length() ) {
                token = token.substring(optID.length());

                StringTokenizer tok = 
                        new StringTokenizer(token, assignmentString);
                if ( tok.hasMoreTokens() ) {
                    arg = tok.nextToken();

//optarg : should be good
                } else if ( ! opt.isArgOptional() ) {
                    throw new OptionException("option '" + token + 
                            "' requires an argument", 
                            OptionException.MISSING_ARGUMENT);
                }

/*
 * FIXME
 * if arg optional then optarg might actually be next option, can't tell
 * if not an arg, don't actually want to take it out of the iterator

    options
        1- remove then push back if necessary
        2- switch from iterator to a index/list arrangement
 */
            /*
             * if the argument is not optional
             */
            } else if ( ! opt.isArgOptional() ) {
                if ( getTokenListIterator().hasNext() ) {
                    arg = getTokenListIterator().next();

                } else {
                    throw new OptionException("option '" + token + 
                            "' requires an argument", 
                            OptionException.MISSING_ARGUMENT);
                }

            } else if ( opt.isArgOptional() ) {
                if ( getTokenListIterator().hasNext() ) {
                    arg = getTokenListIterator().next();

                    /*
                     * check to make sure this is not an option 
                     * if it is, step the token iterator back one
                     */
                    if ( isOption((String) arg) ) {
                        getTokenListIterator().previous();
                        arg = null;
                    }

                }

            } else {
                System.err.println("DefaultGetOptions.getArg():" +
                        " THIS SHOULD NEVER HAPPEN");
            }

        } else if ( opt.getArgType() == Option.BOOLEAN_ARG ) {
            /*
             * see if there are a pair of options blah and noblah
             */
            Option antiOption;
//            String optKey = ( optionName.startsWith(getNegationPrefix()) )
//                            ? optionName.substring(getNegationPrefix().length())
//                            : optionName;
//            if ( opt.getName().startsWith(getNegationPrefix()) ) {
            if ( optID.startsWith(getNegationPrefix()) ) {
                antiOption = get(optID.substring(getNegationPrefix().length()));
                opt.setArg(Boolean.FALSE);

            } else {
                antiOption = get(getNegationPrefix() + opt.getName());
                opt.setArg(Boolean.TRUE);
            }

        }

        if ( ( hasArg ) && ( arg != null ) ) {
            /*
             * arg should be a String, every token is
             */
            arg = processArgs(opt, (String) arg);
            opt.setArg(arg);
        }

        return arg;
    }

/*
    protected boolean isParsedOption(String s) {
        boolean isParsedOption = false;
        Option opt = null
        Iterator i = parsedOptionList.iterator();
        while ( i.hasNext() && ( ! isParsedOption ) ) {
            isParsedOption = ((Option) i.next()).
        }

        return isParsedOption;
    }
*/

    /**
     * Returns whether the given string matches the identifier used to
     * represent the standard input.
     *
     * @param s the string
     * @return whether the string represents standard in
     */
    protected boolean isStdIn(String s) {
        return stdioString.equals(s);
    }

    /**
     * Splits a token into an option name.
     *
     * @param s the token string
     * @return a <code>String</code> containing the option name
     */
    protected String stripOption(String s) {
        String str = stripPrefix(s);
        str = stripAssignment(str);
        return str;
    }

    /**
     * Strips the short or long option prefix off of the token string.
     *
     * @param s the token string
     * @return the string stripped of the option prefix
     */
    protected String stripPrefix(String s) {
        int sLength = s.length();

//        boolean isLong = longOptionsEnabled &&
//                         preferLongOptions &&
//                         s.startsWith(longOptionPrefix) &&
//                         ( sLength > longOptionPrefix.length() );

        boolean isLong = preferLongOptions && isLongOption(s);

//        boolean isShort = ( ! isLong ) &&
//                          s.startsWith(shortOptionPrefix) &&
//                          ( sLength > shortOptionPrefix.length() );

        boolean isShort = ( ! isLong ) && isShortOption(s);

        String strippedOption = "";
        if ( stdioString.equals(s) ) {
            strippedOption = s;

        } else if ( isLong ) {
            strippedOption = s.substring(longOptionPrefix.length());

        } else if ( isShort ) {
            strippedOption = s.substring(shortOptionPrefix.length());
        }

        return strippedOption;
    }

    /**
     * Returns the argument from a token.
     *
     * @param s the token string
     * @return the argument
     */
    protected String stripAssignment(String s) {
        String str = s;

        if ( allowExplicitAssignments ) {
            int index = str.indexOf(assignmentString);

            if ( index != -1 ) {
                int length = str.length() - index;
                str = str.substring(0, index);
            }
        }

        return str;
    }

    /**
     * Throws an <code>OptionException</code>.
     *
     * @throws the exception
     */
    protected void throwOptionException() throws OptionException {
        throw new OptionException();
    }

    /**
     * Throws an <code>OptionException</code>.
     *
     * @param e an <code>OptionException</code>.
     * @throws the exception
     */
    protected void throwOptionException(String errorMsg, int errorCode) 
            throws OptionException {

        throw new OptionException(errorMsg, errorCode);
    }

    /**
     * Returns the usage string for this set of options.
     *
     * @return the usage string
     */
    public String getUsage() {
        final String NEWLINE = System.getProperty("line.separator", "\n");

        StringBuffer buffer = new StringBuffer();

        Iterator i = iterator();

        return buffer.toString();
    }

    /**
     * Returns the usage string for this set of options.
     *
     * @return the usage string
     */
    protected List getParsedOptions() {
        Map uniqueMap = getUniqueMap(optionsMap);

        List parsedOptions = createList();

        OptionRecord or;
        Iterator it = uniqueMap.values().iterator();
        while ( it.hasNext() ) {
            or = (OptionRecord) it.next();
            if ( or.isParsed() ) {
                Number reqIndex = or.getRequestedIndex();
                int index = ( reqIndex != null ) ? reqIndex.intValue() 
                                                 : parsedOptions.size();
                if ( index > parsedOptions.size() ) {
                    index = parsedOptions.size();
                }
                parsedOptions.add(index, or.getOption());
            }
        }

        return parsedOptions;
    }

    /**
     * Returns a <code>Map</code> object containing only the unique values
     * from the given <code>Map</code> object.
     *
     * @param map the given <code>Map</code> object
     * @return the <code>Map</code> object containing only the unique values
     */
    protected static Map getUniqueMap(Map map) {
        Map uniqueMap = new HashMap();

        Object obj;
        Iterator i = map.values().iterator();
        while ( i.hasNext() ) {
            obj = i.next();
            uniqueMap.put(obj, obj);
        }

        return uniqueMap;
    }

    /**
     * Used to encapuslate <code>Option</code> objects.
     *
     * @author Joshua T. Rowe
     * @version 1.0  AUG 27 1999  Initial Version.
     * @JDKVersion 1.2
     */
    protected class OptionRecord extends Object {

//        /**
//         * the Abbreviation property.
//         *
//         * @serial
//         */
//        private String abbreviation;
        
        /**
         * the Parsed Count property.
         *
         * @serial
         */
        private int parsedCount;
        
        /**
         * the Requested Index property.
         *
         * @serial
         */
        private Number requestedIndex;
        
        /**
         * the Actual Index property.
         *
         * @serial
         */
        private Number actualIndex;
        
        /**
         * the Arg property.
         *
         * @serial
         */
        private Object arg;
        
        /**
         * the Option property.
         *
         * @serial
         */
        private Option option;
      
        /**
         * Creates a new <code>OptionRecord</code> object.
         */
        public OptionRecord() {
            super();
            setParsedCount(0);
            setRequestedIndex(null);
            setActualIndex(null);
        }

        /**
         * Sets the Option property.
         *
         * @param option the new value for the Option property
         */
        public void setOption(Option option) {
            this.option = option;
        }

        /**
         * Returns the Option property.
         *
         * @returns the Option property
         */
        public Option getOption() {
            if ( ( option != null ) && 
                 ( option.getArgType() == Option.COUNT_ARG ) ) {

                option.setArg(new Integer(parsedCount));
            }

            return option;
        }

        /**
         * Sets the Arg property.
         *
         * @param newArg the new value for the Arg property
         */
        public void setArg(Object newArg) {
            arg = newArg;
        }

        /**
         * Returns the Arg property.
         *
         * @returns the Arg property
         */
        public Object getArg() {
            return arg;
        }

        /**
         * Sets the Actual Index property.
         *
         * @param actualIndex the new value for the Actual Index property
         */
        public void setActualIndex(int actualIndex) {
            setActualIndex(new Integer(actualIndex));
        }

        /**
         * Sets the Actual Index property.
         *
         * @param actualIndex the new value for the Actual Index property
         */
        public void setActualIndex(Number actualIndex) {
            this.actualIndex = actualIndex;
        }

        /**
         * Returns the Actual Index property.
         *
         * @returns the Actual Index property
         */
        public Number getActualIndex() {
            return this.actualIndex;
        }

        /**
         * Sets the Requested Index property.
         *
         * @param requestedIndex the new value for the Requested Index property
         */
        public void setRequestedIndex(int requestedIndex) {
            setRequestedIndex(new Integer(requestedIndex));
        }

        /**
         * Sets the Requested Index property.
         *
         * @param requestedIndex the new value for the Requested Index property
         */
        public void setRequestedIndex(Number requestedIndex) {
            if ( ( requestedIndex != null ) && 
                 ( requestedIndex.longValue() > 0 ) ) {

                this.requestedIndex = requestedIndex;

            } else {
                this.requestedIndex = null;
            }
        }

        /**
         * Returns the Requested Index property.
         *
         * @returns the Requested Index property
         */
        public Number getRequestedIndex() {
            return requestedIndex;
        }

        /**
         * Sets the Parsed Count property.
         *
         * @param newParsed the new value for the Parsed Count property
         */
        private void setParsedCount(int newParsedCount) {
            parsedCount = newParsedCount;

        }

        /**
         * Sets the Parsed Count property.
         *
         * @param parsed the new value for the Parsed Count property
         */
        public void incrementParsedCount() {
            setParsedCount(getParsedCount() + 1);
        }

        /**
         * Returns the Parsed property.
         *
         * @returns the Parsed Count property
         */
        public int getParsedCount() {
            return this.parsedCount;
        }

        /**
         * Sets the Parsed property.
         *
         * @param parsed the new value for the Parsed property
         */
        public void setParsed(boolean parsed) {
//FIXME : once this stuff is int Option class
//            setParsedCount(getOption().getMaxInstances());
            setParsedCount(((DefaultOption) getOption()).getMaxInstances());
        }

        /**
         * Returns the Parsed property.
         *
         * @returns the Parsed property
         */
        public boolean isParsed() {
//FIXME : once this stuff is int Option class
//            return getParsedCount() >= getOption().getMaxInstances();
            return getParsedCount() >= ((DefaultOption) getOption()).getMaxInstances();
        }

//        /**
//         * Sets the Abbreviation property.
//         *
//         * @param abbreviation the new value for the Abbreviation property
//         */
//        public void setAbbreviation(String abbreviation) {
//            this.abbreviation = abbreviation;
//        }
//
//        /**
//         * Returns the Abbreviation property.
//         *
//         * @returns the Abbreviation property
//         */
//        public String getAbbreviation() {
//            return this.abbreviation;
//        }

        public String toString() {
            StringBuffer buf = new StringBuffer();

            buf.append(getClass().getName());
            buf.append("[");
            buf.append("option=");
            buf.append(getOption());
            buf.append(",");
            buf.append("parsedCount=");
            buf.append(getParsedCount());
            buf.append(",");
            buf.append("]");

            return buf.toString();
        }
    }

    private class OptionsMap {

        public OptionsMap() {
            super();

            map = new HashMap();
        }

        /**
         * the Map property.
         *
         * @serial
         */
        private Map map;
        
        /**
         * Returns the Map property.
         *
         * @returns the Map property
         */
        public Map getMap() {
            return this.map;
        }

        public Set getSet(Object key) {
            return (Set) getMap().get(key);
        }

        public OptionRecord[] get(Object key) {
            Set s = getSet(key);
            OptionRecord[] or = new OptionRecord[0];
            return ( s.size() > 0 ) ? (OptionRecord[]) s.toArray(or) : or;
        }

        public void put(Object key, OptionRecord value) {
            Set s = getSet(key);
            s.add(value);
        }

        public Iterator iterator(Object key) {
            return getSet(key).iterator();
        }
    
    }

}

