/*
 *  DefaultOption.java
 *  Copyright (C) 1999,2000  Joshua T. Rowe
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  To contact the author:
 *  email : jrowe3@home.com
 *  web   : http://jrowe.dhs.org/~jrowe/
 */

package jrowe.getopt;

import java.util.Iterator;

import jrowe.getopt.Option;

/**
 * This is the default implementation of an <code>Option</code>.
 *
 * @author Joshua T. Rowe
 * @version 1.3  MAR 10 2000  Moved from package org.jrowe.getopt to
 *          jrowe.getopt.
 * @version 1.2  FEB 17 2000  Deprecated setGetOptions.  There seems to be
 *          no reason for the <code>Option</code> to know which
 *          <code>GetOptions</code> object it belongs to.
 * @version 1.1  JAN 20 2000  Preparation for release 0.1
 * @version 1.0  AUG 27 1999  Initial documented version.
 */
public class DefaultOption implements Option {

    /**
     * The default long name for <code>Option</code> objects.
     *
     * @serial
     */
    private static final String DEFAULT_NAME = "";

    /**
     * The default short name for <code>Option</code> objects.
     *
     * @serial
     */
    private static final char DEFAULT_CHAR = '\0';

    /**
     * The name of the option.
     * <p>
     *
     * This is the same as the long option argument on the command line.
     *
     * @serial
     */
    private String name;

    /**
     * The short name of the option.
     *
     * @serial
     */
    private char c;

    /**
     * The argument for this <code>Option</code>.
     *
     * @serial
     */
    private Object arg;

    /**
     * Indicates the type of argument this option expects.
     * <p>
     *
     * @serial
     *
     * @see Option#setOption(int)
     * @see Option#getOption()
     */
    private int argType;

    /**
     * The <code>Runnable</code> object to execute when this option is found
     * on the command line.
     *
     * @serial
     */
    private OptionMethod method;

    /**
     * The <code>Runnable</code> object to execute when there is an error with
     * this option on the command line.
     *
     * @serial
     */
    private Runnable errorMethod;

    /**
     * The <code>GetOptions</code> object to which this object belongs.
     *
     * @serial
     * @deprecated
     */
    private GetOptions getOptions;
   
    /**
     * the Min Instances property.
     *
     * @serial
     */
    private int minInstances;
    
    /**
     * the Max Instances property.
     *
     * @serial
     */
    private int maxInstances;
    
    /**
     * the Instance Count property.
     *
     * @serial
     */
    private int instanceCount;

    /**
     * whether the argument for this <code>Option</code> is optional
     *
     * @serial
     */
    private boolean argOptional;

    /**
     * Creates a new <code>DefaultOption</code> with the default attributes.
     */
    public DefaultOption() {    
        super();
        clear();
    }

    /**
     * Reinitializes this <code>Option</code> to it's default state.
     */
    public void clear() {
        setArgType(NO_ARG);
        setArgOptional(true);
        setName(DEFAULT_NAME);
        setChar(DEFAULT_CHAR);
        setArg(null);
        resetInstanceCount();
        setMinInstances(1);
        setMaxInstances(1);
    }

//better way to do this?
//    /**
//     * Sets the <code>GetOptions</code> to which this object belongs.
//     *
//     * @param getOptions the <code>GetOptions</code>
//     * @deprecated
//     */
//    public void setGetOptions(GetOptions getOptions) {
//        this.getOptions = getOptions;
//    }

    /**
     * Sets the long name of this object.
     *
     * @param name the long name
     */
    public void setName(String name) {
        this.name = name;
    }

    /**
     * Gets the long name of this object.
     *
     * @return the long name.
     */
    public String getName() {
        return name;
    }

    /**
     * Sets the argument for this program option.
     *
     * @param newArg the argument for this program option
     */
    public void setArg(Object newArg) {
        arg = newArg;
    }
    
    /**
     * Returns the argument for this program option.
     *
     * @return the argument for this program option
     */
    public Object getArg() {
//FIXME : use instances if argtype == COUNT_ARG
        return arg;
    }

    /**
     * Sets the argument type for this program option.
     *
     * @param argType The type of argument this program option takes.<br>
     *        Valid types are:<br>
     *        <code>NO_ARG</code><br>
     *        <code>SCALAR_ARG</code><br>
     *        <code>STRING_ARG</code><br>
     *        <code>LIST_ARG</code><br>
     *        <code>COUNT_ARG</code><br>
     *        <code>NUMERIC_ARG</code><br>
     *        <code>REAL_ARG</code><br>
     *        <code>INTEGER_ARG</code><br>
     *        <code>BOOLEAN_ARG</code><br>
     *        <!-- <code>NESTED_ARG</code><br> -->
     */
    public void setArgType(int argType) {
        this.argType = argType;

        switch ( argType ) {
        case COUNT_ARG :
//FIXME : use instances
            setArg(new Long(0));
            setArgOptional(false);
            setMaxInstances(Integer.MAX_VALUE);
            break;

        default :
            break;
        }
    }
    
    /**
     * Returns the argument type for this program option.
     *
     * @return the argument type for this program option
     *
     * @see #setArgType(int)
     */
    public int getArgType() {
        return argType;
    }

    /**
     * Returns whether the given argument type has an argument or not.
     *
     *
     * @param argType the given arg type
     * @return whether the given argument type has an argument or not.
     */
    public static boolean hasArg(int argType) {
//ENHANCE
        return ( argType != NO_ARG ) && 
               ( argType != COUNT_ARG ) &&
               ( argType != BOOLEAN_ARG );
    }

    /**
     * Sets whether the argument for this <code>Option</code> is
     * optional
     *
     * @return argOptional whether the argument for this <code>Option</code> 
     *         is optional
     */
    public void setArgOptional(boolean argOptional) {
        this.argOptional = argOptional;
    }

    /**
     * Returns whether the argument for this <code>Option</code> is
     * optional
     *
     * @return whether the argument for this <code>Option</code> is
     *         optional
     */
    public boolean isArgOptional() {
        return argOptional;
    }

//    public void setHasArg(boolean hasArg) {
//        this.hasArg = hasArg;
//    }
//
//    public boolean getHasArg() {
//        return hasArg;
//    }
//
//    public void setArgCount(int argCount) {
//        this.argCount = argCount;
//    }
//
//    public int getArgCount() {
//        return argCount;
//    }
//
    /**
     * Sets the short name of this program option.
     *
     * @param c the short name of this program option
     */
    public void setChar(char c) {
        this.c = c;
    }

    /**
     * Returns the short name of this program option.
     *
     * @return the short name of this program option
     */
    public char getChar() {
        return c;
    }

    /**
     * Sets the short name of this program option.
     *
     * @param c the short name of this program option
     */
    public void setCharacter(Character c) {
        this.c = c.charValue();
    }

    /**
     * Returns the short name of this program option.
     *
     * @return the short name of this program option
     */
    public Character getCharacter() {
        return new Character(c);
    }

    /**
     * Sets the <code>Runnable</code> object to invoke if the program 
     * option is found.
     *
     * @param method the object to invoke
     */
    public void setMethod(OptionMethod method) {
        this.method = method;
    }

    /**
     * Invokes the <code>Runnable</code> object that is to be run if this
     * program option is found.
     */
    public void runMethod() {
        if ( method != null ) {
            method.run();
        }
    }
    
    /**
     * Invokes the <code>Runnable</code> object that is to be run if this
     * program option is found.
     *
     * @param arg the argument
     * @deprecated
     */
    public void runMethod(Object arg) {
        setArg(arg);
        runMethod();
    }
 
//    public void setMethod(Method method);

    public void setErrorMethod(Runnable errorMethod) {
        this.errorMethod = errorMethod;
    }

//    public void setErrorMethod(Method method);

//    public void parseOption(String optionString) {
//    }
//
//    public String toOptionString() {
//        return this.toString();
//    }

    /**
     * Resets the instance count of this <code>Option</code>.
     */
    public void resetInstanceCount() {
        instanceCount = 0;
    }

    /**
     * Increments the instance count of this <code>Option</code>.
     *
     * @return the new value of the instance count property
     */
    public int incrementInstanceCount() {
        return ++instanceCount;
    }

//Doesn't make sense unless
//    /**
//     * Returns the number of times this <code>Option</code> appears on
//     * the command line.
//     *
//     * @return the number of times this <code>Option</code> appears on
//     *         the command line
//     */
//    public int getInstanceCount() {
//        return instanceCount;
//    }

    /**
     * Sets the minimum instances of this option that must appear if any
     * appear at all.
     * <p>
     *
     * In other words, if this is set to <code>N</code>, there must either be
     * <code>0</code> instances of this <code>Option</code> or there must be
     * at least <code>3</code> instances of this </code>Option</code>.
     * <p>
     *
     * @param minInstances the minimum number of instances.<br>
     *        If this is less than <code>1</code>, an
     *        <code>IllegalArgumentException</code> will be thrown.<br>
     *        If this is greater than the <code>maxInstances</code>, property 
     *        an <code>IllegalArgumentException</code> will be thrown.<br>
     */
    public void setMinInstances(int minInstances) {
        this.minInstances = minInstances;
    }

    /**
     * Returns the minimum instances of this option that must appear if any
     * appear at all.
     * <p>
     *
     * In other words, if this is set to <code>N</code>, there must either be
     * <code>0</code> instances of this <code>Option</code> or there must be
     * at least <code>3</code> instances of this </code>Option</code>.
     * <p>
     *
     * @return the minimum number of instances.<br>
     *
     * @see #setMinInstances(int)
     */
    public int getMinInstances() {
        return minInstances;
    }

    /**
     * Sets the maximum instances of this option that may appear.
     * <p>
     *
     * @param maxInstances the maximum number of instances.<br>
     *        If this is less than the <code>minInstances</code> property, an
     *        <code>IllegalArgumentException</code> will be thrown.
     */
    public void setMaxInstances(int maxInstances) {
//        switch ( getArgType() ) {
//        case COUNT_ARG :
//            break;
//
//        default :
            this.maxInstances = maxInstances;
//            break;
//        }
    }

    /**
     * Returns the maximum instances of this option that must appear if any
     * appear at all.
     * <p>
     *
     * @return the maximum number of instances.<br>
     */
    public int getMaxInstances() {
        return maxInstances;
    }

    /**
     * Creates and returns a copy of this object.
     * <p>
     *
     * @exception  OutOfMemoryError if there is not enough memory.
     *
     * @see        java.lang.Cloneable
     * @see        java.lang.Object#clone()
     */
    public Object clone() {
        Object clone = null;

        try {
            clone = super.clone();

        } catch (CloneNotSupportedException e) {
            e.printStackTrace();
        }

        return clone;
    }

    public String toString() {
        StringBuffer buf = new StringBuffer();

        buf.append(getClass().getName());
        buf.append("[");
        buf.append("name=");
        buf.append(getName());
        buf.append(",");
        buf.append("character=");
        buf.append(getCharacter());
        buf.append(",");
        buf.append("argType=");
        buf.append(getArgType());
        buf.append(",");
        buf.append("argOptional=");
        buf.append(isArgOptional());
        buf.append(",");
        buf.append("arg=");
        buf.append(getArg());
        buf.append("]");

        return buf.toString();
    }

    /**
     * Creates an <code>Option</code> of the same class as this 
     * <code>Option</code>.
     * <p>
     *
     * This is similar to the <code>clone</code> method except none of the
     * data contained in the object is copied.
     * <p>
     *
     * @see #clone()
     */
    public Option createNewInstance() {
        Option opt = (Option) this.clone();
        opt.clear();
        return opt;
    }

}

