/*
 *  Option.java
 *  Copyright (C) 1999,2000  Joshua T. Rowe
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  To contact the author:
 *  email : jrowe3@home.com
 *  web   : http://jrowe.dhs.org/~jrowe/
 */

package jrowe.getopt;

import java.util.Iterator;

/**
 * An interface representing a program option.
 *
 * @author Joshua T. Rowe
 * @version 1.2  MAR 10 2000  Moved from package org.jrowe.getopt to
 *          jrowe.getopt.
 * @version 1.1  FEB 17 2000  Deprecated setGetOptions.  There seems to be
 *          no reason for the <code>Option</code> to know which
 *          <code>GetOptions</code> object it belongs to.
 * @version 1.0  JAN 21 2000  Initial documented version.  Preparation for
 *          release 0.1
 */
public interface Option extends Cloneable {

    /**
     * An indicator used when the program option takes no arguments.
     *
     * @serial
     */
    public static final int NO_ARG = 0;

    /**
     * An indicator used when the program option takes a scalar argument.
     *
     * @serial
     */
    public static final int SCALAR_ARG = 99;

    /**
     * An indicator used when the program option takes a string argument.
     *
     * @serial
     */
    public static final int STRING_ARG = 1;

    /**
     * <p>
     * An indicator used when the program option takes an integer argument.
     * </p>
     *
     * <p>
     * NOTE:<br>
     * <blockquote>
     * Integer being defined in the mathematical sense, not strictly an
     * <code>Integer</code> type in Java.
     * </blockquote>
     * </p>
     *
     * @serial
     */
    public static final int INTEGER_ARG = 2;

    /**
     * <p>
     * An indicator used when the program option takes an real argument.
     * </p>
     *
     * @serial
     */
    public static final int REAL_ARG = 3;

    /**
     * An indicator used when the program option takes a list of arguments.
     *
     * @serial
     */
    public static final int LIST_ARG = 4;

    /*
     * An indicator used when the program option consists of more program
     * options.
     *
     * @serial
     */
    public static final int NESTED_ARG = 5;

    /**
     * An indicator used when the program option takes a boolean argument.
     *
     * @serial
     */
    public static final int BOOLEAN_ARG = 6;

    /**
     * An indicator used when the program option takes as it's argument
     * the number of times the option appears on the command line.
     *
     * @serial
     */
    public static final int COUNT_ARG = 7;

    /**
     * An indicator used when the program option takes a numeric argument.
     *
     * @serial
     */
    public static final int NUMERIC_ARG = 8;

    /** 
     * Sets the long name for this program option.
     *
     * @param name the name for this option
     */
    public void setName(String name);

    /** 
     * Returns the long name for this program option.
     *
     * @return the name for this option
     */
    public String getName();

    /**
     * Sets the argument for this program option.
     *
     * @param arg the argument for this program option
     */
    public void setArg(Object arg);
    
    /**
     * Returns the argument for this program option.
     *
     * @return the argument for this program option
     */
    public Object getArg();

    /**
     * Sets the argument type for this program option.
     *
     * @param argType The type of argument this program option takes.<br>
     *        Valid types are:<br>
     *        <code>NO_ARG</code><br>
     *        <code>SCALAR_ARG</code><br>
     *        <code>STRING_ARG</code><br>
     *        <code>LIST_ARG</code><br>
     *        <code>COUNT_ARG</code><br>
     *        <code>NUMERIC_ARG</code><br>
     *        <code>REAL_ARG</code><br>
     *        <code>INTEGER_ARG</code><br>
     *        <code>BOOLEAN_ARG</code><br>
     *        <!-- <code>NESTED_ARG</code><br> -->
     */
    public void setArgType(int argType);

    /**
     * Returns the argument type for this program option.
     *
     * @return the argument type for this program option
     *
     * @see #setArgType(int)
     */
    public int getArgType();

    /**
     * Sets whether the argument for this <code>Option</code> is
     * optional
     *
     * @return argOptional whether the argument for this <code>Option</code> 
     *         is optional
     */
    public void setArgOptional(boolean argOptional);

    /**
     * Returns whether the argument for this <code>Option</code> is
     * optional
     *
     * @return whether the argument for this <code>Option</code> is
     *         optional
     */
    public boolean isArgOptional();

//    public void setMinInstances(int minInstances);

//    public int getMinInstances();

//    public void setMaxInstances(int maxInstances);

//    public int getMaxInstances();

    /**
     * Reinitializes this <code>Option</code> to it's default state.
     */
    public void clear();

    /**
     * Sets the short name of this program option.
     *
     * @param c the short name of this program option
     */
    public void setChar(char c);

    /**
     * Returns the short name of this program option.
     *
     * @return the short name of this program option
     */
    public char getChar();

    /**
     * Sets the short name of this program option.
     *
     * @param c the short name of this program option
     */
    public void setCharacter(Character c);

    /**
     * Returns the short name of this program option.
     *
     * @return the short name of this program option
     */
    public Character getCharacter();

    /**
     * Sets the <code>Runnable</code> object to invoke if the program 
     * option is found.
     *
     * @param method the object to invoke
     */
    public void setMethod(OptionMethod method);
    
    /**
     * Invokes the <code>Runnable</code> object that is to be run if this
     * program option is found.
     */
    public void runMethod();

    /**
     * Invokes the <code>Runnable</code> object that is to be run if this
     * program option is found.
     *
     * @param arg the argument
     * @deprecated
     */
    public void runMethod(Object arg);

//    public void setErrorMethod(Runnable errorMethod);

//    public void setErrorMethod(Method errorMethod);

//    public void parseOption(String optionString);

//    public String toOptionString();

//    public String toUsageString();

//better way to do this?
//    /**
//     * Sets the <code>GetOptions</code> to which this object belongs.
//     *
//     * @param getOptions the <code>GetOptions</code>
//     * @deprecated
//     */
//    public void setGetOptions(GetOptions getOptions);
    
    /**
     * Creates and returns a copy of this object.
     * <p>
     *
     * @exception  OutOfMemoryError            if there is not enough memory.
     * @see        java.lang.Cloneable
     * @see        java.lang.Object#clone()
     */
    public Object clone();

    /**
     * Creates an <code>Option</code> of the same class as this 
     * <code>Option</code>.
     * <p>
     *
     * This is similar to the <code>clone</code> method except none of the
     * data contained in the object is copied.
     * <p>
     *
     * @see #clone()
     */
    public Option createNewInstance();

}

