/*
 *  Tester.java
 *  Copyright (C) 2000  Joshua T. Rowe
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  To contact the author:
 *  email : jrowe3@home.com
 *  web   : http://jrowe.dhs.org/~jrowe/
 */

package jrowe.test;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;

import java.io.Serializable;

import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import junit.framework.Test;
import junit.framework.TestSuite;

import junit.util.StandardTestSuiteLoader;
import junit.util.TestSuiteLoader;

/**
 * @author Joshua T. Rowe
 * @version 1.0  30 MAY 2000  Intial version.
 */
public class Tester extends Object implements Serializable {

    /**
     * The name of the system property which stores the gui indicator.
     * The gui indicator indicates what form of user interface to use.
     *
     * @serial
     */
    public static final String UI_PROP = Tester.class.getName() + ":junit.ui";

    /**
     * Indicates the user interface ought to use the AWT widget set.
     *
     * @serial
     */
    public static final String NATIVE_PROP = "native";

    /**
     * Indicates the user interface ought to use the Swing widget set.
     *
     * @serial
     */
    public static final String SWING_PROP = "swing";

    /**
     * Indicates the user interface ought to use the command line.
     *
     * @serial
     */
    public static final String CONSOLE_PROP = "console";

    /**
     * Creates a <code>Tester</code>.
     */
    public Tester() {
        super();
    }

    private TestSuiteLoader createDefaultTestSuiteLoader() {
        return new StandardTestSuiteLoader();
    }

    /**
     * Sets this <code>Tester</code> to use the console based user interface.
     */
    public void setConsole() {
        System.setProperty(UI_PROP, CONSOLE_PROP);
    }

    /**
     * Returns whether the console based user interface will be used or not.
     *
     * @returns whether the console based user interface will be used or not.
     */
    public boolean isConsole() {
        return CONSOLE_PROP.equalsIgnoreCase(System.getProperty(UI_PROP));
    }

    /**
     * Sets this <code>Tester</code> to use the Swing graphical user interface.
     */
    public void setSwingGUI(boolean newSwingGUI) {
        System.setProperty(UI_PROP, SWING_PROP);
    }

    /**
     * Returns whether the Swing graphical user interface will be used or not.
     *
     * @returns whether the Swing graphical user interface will be used or not.
     */
    public boolean isSwingGUI() {
        return SWING_PROP.equalsIgnoreCase(System.getProperty(UI_PROP));
    }

    /**
     * Sets this <code>Tester</code> to use the AWT graphical user interface.
     */
    public void setNativeGUI() {
        System.setProperty(UI_PROP, NATIVE_PROP);
    }

    /**
     * Returns whether the AWT graphical user interface will be used or not.
     *
     * @returns whether the AWT graphical user interface will be used or not.
     */
    public boolean isNativeGUI() {
        return NATIVE_PROP.equalsIgnoreCase(System.getProperty(UI_PROP));
    }

    /**
     * Creates a <code>Test</code> from the specified class.
     *
     * @param suiteClass the specified class
     * @deprecated
     */
    public static Test createSuite(Class suiteClass) {
        return new TestSuite(suiteClass);
    }

//    /**
//     * @deprecated
//     */
//    public static Test _createSuite(Class suiteClass) {
//        Test suite = null;
//
//        final Class[] EMPTY_ARRAY = new Class[] {};
//        try {
//            Method method = suiteClass.getMethod("suite", EMPTY_ARRAY);
//            suite = (Test) method.invoke(null, EMPTY_ARRAY);
//
//        } catch (SecurityException e) {
//            e.printStackTrace();
//
//        } catch (NoSuchMethodException e) {
////            Class superclass = suiteClass.getSuperclass();
////            System.out.println("NSME:superclass=" + superclass);
////            if ( superclass == null ) {
//                e.printStackTrace();
////
////            } else {
////                suite = createSuite(superclass);
////            }
//
//        } catch (IllegalAccessException e) {
//            e.printStackTrace();
//
//        } catch (IllegalArgumentException e) {
//            e.printStackTrace();
//
//        } catch (InvocationTargetException e) {
//            e.printStackTrace();
//        }
//
//        return suite;
//    }

    public void runSuite(Class cls) {
        runSuite(cls.getName());
    }

    public void runSuite(String suiteName) {
        if ( isNativeGUI() ) {
            runSuiteNative(suiteName);

        } else if ( isSwingGUI() ) {
            runSuiteSwing(suiteName);

        } else {
            runSuiteConsole(suiteName);
        }
    }

    public void runSuite(Test test) {
        if ( test instanceof TestSuite ) {
            runSuite((TestSuite) test);

        } else {
            System.err.println(getClass().getName() + ".runSuite(" +
                    Test.class.getName() + "): test is not an instance of " +
                    TestSuite.class.getName());
        }
    }

    public void runSuite(TestSuite suite) {
//FIXME Apparently cannot use a TestSuite object to load a gui and run it
//        if ( isNativeGUI() ) {
//            runSuiteNative(suite);
//
//        } else if ( isSwingGUI() ) {
//            runSuiteSwing(suite);
//
//        } else {
            runSuiteConsole(suite);
//        }
    }

    public void runSuiteNative(String className) {
        junit.ui.TestRunner testRunner = new junit.ui.TestRunner();
        testRunner.start(new String[] { className }, 
                createDefaultTestSuiteLoader());
    }

    public void runSuiteSwing(String className) {
        junit.swingui.TestRunner testRunner = new junit.swingui.TestRunner();
        testRunner.start(new String[] { className }, 
                createDefaultTestSuiteLoader());
    }

    public void runSuiteConsole(String className) {
        try {
            runSuiteConsole(new TestSuite(Class.forName(className)));

        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }
    }

    public void runSuiteConsole(Test suite) {
        junit.textui.TestRunner.run(suite);
    }

    public static TestSuite createTestSuite(String className) {
        Class cls = null;

        try {
            cls = Class.forName(className);

        } catch (ClassNotFoundException e) {
            e.printStackTrace();
        }

        return createTestSuite(cls);
    }

    public static TestSuite createTestSuite(Class cls) {
        TestSuite suite = null;

        final Class[] NO_ARG_ARRAY = new Class[] {};
        try {
            Method suiteMethod = cls.getMethod("suite", NO_ARG_ARRAY);
            suite = (TestSuite) suiteMethod.invoke(null, NO_ARG_ARRAY);

        } catch (SecurityException e) {
            e.printStackTrace();

        } catch (NoSuchMethodException e) {
            e.printStackTrace();

        } catch (IllegalAccessException e) {
            e.printStackTrace();

        } catch (IllegalArgumentException e) {
            e.printStackTrace();

        } catch (InvocationTargetException e) {
            e.printStackTrace();
        }

        return suite;
    }

    public static void main(String[] args) {
        if ( args.length < 1 ) {
            System.err.println(Tester.class.getName() + ".main(" +
                    String[].class.getName() + "):argument 1 must be a " +
                    "class name.  Doing nothing.");

            System.exit(1);
        }

        TestSuite suite = createTestSuite(args[0]);

        if ( suite != null ) {
            new Tester().runSuite(suite);

        } else {
            System.err.println(Tester.class.getName() + ".main(" +
                    String[].class.getName() + "):unable to create a " +
                    TestSuite.class.getName() + " object.  Doing nothing.");
        }
    }

}

