/*
 *  GetoptTest.java
 *  Copyright (C) 2000  Joshua T. Rowe
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  To contact the author:
 *  email : jrowe3@home.com
 *  web   : http://jrowe.dhs.org/~jrowe/
 */

package jrowe.test.getopt;

import java.lang.reflect.Method;

import java.io.File;

import java.math.BigDecimal;
import java.math.BigInteger;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import jrowe.getopt.AbstractOptionMethod;
import jrowe.getopt.DefaultGetOptions;
import jrowe.getopt.DefaultOption;
import jrowe.getopt.GetOptions;
import jrowe.getopt.Option;
import jrowe.getopt.OptionException;

import jrowe.test.TestCase;

import junit.framework.AssertionFailedError;
import junit.framework.Test;
import junit.framework.TestSuite;

/*
 * @version 1.1 MAR 23 2000  Changed superclass.  Added setUp.
 * @version 1.0 MAR 11 2000  Initial version
 */
public class GetoptTest extends TestCase {

    /*
     * make sure first character of all options are unique or set the
     * Character property of each Option in init by hand
     */
    private static final String NOARG = "version";
    private static final Character NOARG_CHAR = new Character(NOARG.charAt(0));

    private static final String MULT = "mult";
    private static final Character MULT_CHAR = new Character(MULT.charAt(0));

    private static final String SCALAR_ARG = "file";
    private static final Character SCALAR_ARG_CHAR = 
            new Character(SCALAR_ARG.charAt(0));

    private static final String SCALAR_OPTIONAL_ARG = "string";
    private static final Character SCALAR_OPTIONAL_ARG_CHAR = 
            new Character(SCALAR_ARG.charAt(0));

    private static final String BOOLEAN_ARG = "aBoolean";
    private static final Character BOOLEAN_ARG_CHAR = 
            new Character(BOOLEAN_ARG.charAt(0));

    private static final String LIST_ARG = "list";
    private static final Character LIST_ARG_CHAR = 
            new Character(LIST_ARG.charAt(0));

    private static final String INTEGER_ARG = "integer";
    private static final Character INTEGER_ARG_CHAR = 
            new Character(INTEGER_ARG.charAt(0));

    private static final String BIG_INTEGER_ARG = "big-integer";
    private static final Character BIG_INTEGER_ARG_CHAR = 
            new Character(BIG_INTEGER_ARG.charAt(0));

    private static final String REAL_ARG = "real";
    private static final Character REAL_ARG_CHAR = 
            new Character(REAL_ARG.charAt(0));

    private static final String BIG_REAL_ARG = "Big-decimal";
    private static final Character BIG_REAL_ARG_CHAR = 
            new Character(BIG_REAL_ARG.charAt(0));

    private static final String COUNT_ARG = "count";
    private static final Character COUNT_ARG_CHAR = 
            new Character(COUNT_ARG.charAt(0));

    private Map flagMap;
    private GetOptions getOptions;

    public GetoptTest(String name) {
        super(name);
    }

    private String join(List l, String joinStr) {
        StringBuffer buf = new StringBuffer();

        Iterator it = l.iterator();
        if ( it.hasNext() ) {
            buf.append(it.next().toString());
            while ( it.hasNext() ) {
                buf.append(joinStr);
                buf.append(it.next().toString());
            }
        }

        return buf.toString();
    }

    private String join(String[] s) {
        StringBuffer buf = new StringBuffer();

        for ( int i = 0; i < s.length; i++ ) {
            buf.append(s[i]);
        }

        return buf.toString();
    }

    public void setUp() {
        flagMap = new HashMap();
        getOptions = new DefaultGetOptions();

        Option option;

        option = new DefaultOption();
        option.setName(NOARG);
        option.setChar(NOARG_CHAR.charValue());
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

//        option = new DefaultOption();
//        option.setName(MULT);
//        option.setChar(MULT_CHAR.charValue());
//        option.setMethod(new OptionMethod(option) {
//            public void run() {
//                super.run();
//            }
//        });
//        getOptions.add(option);
//        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(SCALAR_ARG);
        option.setChar(SCALAR_ARG_CHAR.charValue());
        option.setArgType(Option.SCALAR_ARG);
        option.setMethod(new OptionMethod(option));
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(SCALAR_OPTIONAL_ARG);
        option.setChar(SCALAR_OPTIONAL_ARG_CHAR.charValue());
        option.setArgType(Option.SCALAR_ARG);
        option.setArgOptional(true);
        option.setMethod(new OptionMethod(option));
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(BOOLEAN_ARG);
        option.setChar(BOOLEAN_ARG_CHAR.charValue());
        option.setArgType(Option.BOOLEAN_ARG);
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(INTEGER_ARG);
        option.setChar(INTEGER_ARG_CHAR.charValue());
        option.setArgType(Option.INTEGER_ARG);
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(BIG_INTEGER_ARG);
        option.setChar(BIG_INTEGER_ARG_CHAR.charValue());
        option.setArgType(Option.INTEGER_ARG);
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(REAL_ARG);
        option.setChar(REAL_ARG_CHAR.charValue());
        option.setArgType(Option.REAL_ARG);
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(BIG_REAL_ARG);
        option.setChar(BIG_REAL_ARG_CHAR.charValue());
        option.setArgType(Option.REAL_ARG);
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(LIST_ARG);
        option.setChar(LIST_ARG_CHAR.charValue());
        option.setArgType(Option.LIST_ARG);
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();

//                List list = (List) getOption().getArg();
//                if ( list != null ) {
//                    Iterator it = list.iterator();
//                    while ( it.hasNext() ) {
//                        System.out.println("list contents: '" + 
//                                it.next().toString() + "'");
//                    }
//
//                } else {
//                    System.out.println("arg for " + getOption().getName() +
//                            " is null");
//                }
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

        option = new DefaultOption();
        option.setName(COUNT_ARG);
        option.setChar(COUNT_ARG_CHAR.charValue());
        option.setArgType(Option.COUNT_ARG);
        option.setMethod(new OptionMethod(option) {
            public void run() {
                super.run();
            }
        });
        getOptions.add(option);
        flagMap.put(option.getName(), Boolean.FALSE);

    }

    /**
     * Returns whether the option was found or not.
     */
    private boolean longOptionCheck(Object obj, String name) {
        Option opt = (Option) obj;
        return flagMap.get(opt.getName()).equals(new Boolean(
            (opt.getName().equals(name))));
    }

    /**
     * Returns whether the option was found or not.
     */
    private boolean shortOptionCheck(Object obj, Character c) {
        Option opt = (Option) obj;
        return flagMap.get(opt.getName()).equals(new Boolean(
            (opt.getCharacter().equals(c))));
    }

    /**
     * Returns whether the specified option was found.
     */
    private boolean checkOptionFound(Object obj) {
        Option opt = (Option) obj;
        return flagMap.get(opt.getName()).equals(Boolean.TRUE);
    }

    public void testRemove1() {
        getOptions.remove(NOARG);

        Option opt = getOptions.get(NOARG);

        try {
            assert(opt == null);

        } catch (AssertionFailedError e) {
            System.out.println("option to be removed was " + NOARG);
            System.out.println("found opt=" + opt);

//            System.out.println("known options:");
//            System.out.println("/known options:");

            throw e;
        }
    }

    public void testNoArgs() {
        String[] str = new String[] {};
        getOptions.setArgs(str);

        boolean b = true;

        Option opt;
        Iterator it = getOptions.iterator();
        while ( ( b ) && ( it.hasNext() ) ) {
            b = ! checkOptionFound(it.next());
        }

        assert(b);
    }

    public void testEmptyString() {
        String[] str = new String[] { "" };
        getOptions.setArgs(str);

        boolean b = true;

        Option opt;
        Iterator it = getOptions.iterator();
        while ( ( b ) && ( it.hasNext() ) ) {
            b = flagMap.get(((Option) it.next()).getName()).
                equals(Boolean.FALSE);
        }

        assert(b);
    }

//FIXME : functionality this tests is broken
    public void _testShortWithLongPrefix() {
        String[] str = new String[] { "--" + NOARG_CHAR };
        getOptions.setArgs(str);

        boolean b = true;

        Option opt;
        Iterator it = getOptions.iterator();
        while ( ( b ) && ( it.hasNext() ) ) {
            opt = (Option) it.next();
            b = ! checkOptionFound(it.next());
        }

        assert(b);
    }

    public void testLongNoArgOption() {
        String[] str = new String[] { "--" + NOARG };
        getOptions.setArgs(str);

        boolean b = true;

        Option opt;
        Iterator it = getOptions.iterator();
        while ( ( b ) && ( it.hasNext() ) ) {
            opt = (Option) it.next();
            b = longOptionCheck(opt, NOARG);
        }

        assert(b);
    }

    public void testShortNoArgOption() {
        String[] str = new String[] { "-" + NOARG_CHAR.toString() };
        getOptions.setArgs(str);

        boolean b = true;

        Option opt;
        Iterator it = getOptions.iterator();
        while ( ( b ) && ( it.hasNext() ) ) {
            opt = (Option) it.next();
            b = flagMap.get(opt.getName()).equals(new Boolean(
                (opt.getCharacter().equals(NOARG_CHAR))));
        }

        assert(b);
    }

    public void testLongScalarOptionalArgOption1() {
        final String ARG = "Makefile";
        String[] str = new String[] { "--" + SCALAR_OPTIONAL_ARG, ARG };
        getOptions.setArgs(str);

        Option opt = getOptions.get(SCALAR_OPTIONAL_ARG);
        try {
            assert(ARG.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongScalarOptionalArgOption2() {
        final String ARG = "--" + NOARG;
        String[] str = new String[] { "--" + SCALAR_OPTIONAL_ARG, ARG };
        getOptions.setArgs(str);

        Option opt = getOptions.get(SCALAR_OPTIONAL_ARG);
        try {
            assert(opt.getArg() == null);

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongScalarOptionalArgOption3() {
        String[] str = new String[] { "--" + SCALAR_OPTIONAL_ARG };
        getOptions.setArgs(str);

        Option opt = getOptions.get(SCALAR_OPTIONAL_ARG);
        try {
            assert(opt.getArg() == null);

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongScalarArgOption1() {
        final String ARG = "Makefile";
        String[] str = new String[] { "--" + SCALAR_ARG, ARG };
        getOptions.setArgs(str);

        Option opt = getOptions.get(SCALAR_ARG);
        try {
            assert(ARG.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongScalarArgOption2() {
        final String ARG = "Makefile";
        String[] str = new String[] { "--" + SCALAR_ARG + " " + ARG };
        getOptions.setArgs(str);

        Option opt = getOptions.get(SCALAR_ARG);
        try {
            assert(ARG.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    /**
     * Tests to see if the scalar_arg can take an argument with 
     * double quotes.
     */
    public void _testLongScalarArgOption3() {
        final String ARG = "Makefile";
        final String Q = "\"";
        String[] str = new String[] { "--" + SCALAR_ARG + " " + Q + ARG + Q };
        getOptions.setArgs(str);

        Option opt = getOptions.get(SCALAR_ARG);
        try {
            assert(ARG.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("cmd line=\"" + join(str) + "\"");
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    /**
     * Tests to see if the scalar_arg can take an argument with 
     * double quotes.
     */
    public void _testLongScalarArgOption4() {
        final String ARG = "Makefile";
        final String Q = "'";
        String[] str = new String[] { "--" + SCALAR_ARG + " " + Q + ARG + Q };
        getOptions.setArgs(str);

        Option opt = getOptions.get(SCALAR_ARG);
        try {
            assert(ARG.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("cmd line=\"" + join(str) + "\"");
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongBooleanArgOption1() {
        String[] str = new String[] { "--" + BOOLEAN_ARG };
        getOptions.setArgs(str);

        Option opt = getOptions.get(BOOLEAN_ARG);
        try {
            assert(Boolean.TRUE.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("opt.getArg().getClass()='" + 
                    opt.getArg().getClass() + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    /**
     * Tests to see if a long boolean option can be negated by prefixing
     * "no" to the option name
     */
    public void testLongBooleanArgOption2() {
//not a robust test "no" is hardcoded
        final String ARG_NAME = "no" + BOOLEAN_ARG;
        String[] str = new String[] { "--" + ARG_NAME };
        getOptions.setArgs(str);

        Option opt = getOptions.get(ARG_NAME);
        try {
            assert(Boolean.FALSE.equals(opt.getArg()));

        } catch (NullPointerException e) {
            System.err.println();
            System.err.println("NullPointerException");
            System.err.println("opt='" + opt + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    //isn't this the same as above?
    public void __testLongBooleanArgOption2() {
        final String ARG = "not_a_boolean";
        String[] str = new String[] { "--" + BOOLEAN_ARG, ARG };

        try {
            try {
                getOptions.setArgs(str);

            } catch (OptionException e) {
                assert(e.getErrorCode() == OptionException.MALFORMED_ARGUMENT);
            }

        } catch (AssertionFailedError e) {
            throw e;
        }
    }

    public void testLongListArgOption1() {
        final String ARG = "a,b,c,d";
        String[] str = new String[] { "--" + LIST_ARG + " " + ARG };
        getOptions.setArgs(str);

        Option opt = getOptions.get(LIST_ARG);
        try {
            String reconstructedArg = join((List) opt.getArg(), ",");
            assert(ARG.equals(reconstructedArg));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongIntegerArgOption() {
        final Number ARG = new Integer(0);
        String[] str = new String[] { "--" + INTEGER_ARG, ARG.toString() };
        getOptions.setArgs(str);

        Option opt = getOptions.get(INTEGER_ARG);
        try {
            assert(ARG.longValue() == ((Number) opt.getArg()).longValue());

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongBigIntegerArgOption() {
        /* force it to be larger than is representable by a long */
        BigInteger ARG = new BigInteger(Long.toString(Long.MAX_VALUE));
        ARG = ARG.add(BigInteger.ONE);

        String[] str = new String[] { "--" + BIG_INTEGER_ARG, ARG.toString() };
        getOptions.setArgs(str);

        Option opt = getOptions.get(BIG_INTEGER_ARG);
        try {
            assert(ARG.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongRealArgOption() {
        final Number ARG = new Double(0);
        String[] str = new String[] { "--" + REAL_ARG, ARG.toString() };
        getOptions.setArgs(str);

        Option opt = getOptions.get(REAL_ARG);
        try {
//FIXME : very dangerous to compare for equality
            assert(ARG.equals(opt.getArg()));
            //assert(ARG.doubleValue() == ((Number) opt.getArg()).doubleValue());

        } catch (AssertionFailedError e) {
        } catch (NullPointerException e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt='" + opt + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");

            throw e;
        }
    }

    public void testLongBigRealArgOption() {
        /* force it to be larger than is representable by a double */
        BigDecimal ARG = new BigDecimal(Double.MAX_VALUE);
        ARG = ARG.multiply(new BigDecimal(Double.MAX_VALUE));

        String[] str = new String[] { "--" + BIG_REAL_ARG, ARG.toString() };
        getOptions.setArgs(str);

        Option opt = getOptions.get(BIG_REAL_ARG);
        try {
//FIXME : very dangerous to compare for equality
            assert(ARG.equals(opt.getArg()));

        } catch (AssertionFailedError e) {
            System.err.println();
            System.err.println("ARG='" + ARG + "'");
            System.err.println("opt.getArg()='" + opt.getArg() + "'");
            System.err.println("opt.getArg().getClass()=" + 
                    opt.getArg().getClass());

            throw e;
        }
    }

    public void testLongCount0ArgOption() {
        int arg = 0;
        StringBuffer argBuf = new StringBuffer();
        for ( int i = 0; i < arg; i++ ) {
            argBuf.append("--");
            argBuf.append(COUNT_ARG);
            argBuf.append(" ");
        }

        String[] str = new String[] { argBuf.toString() };
        getOptions.setArgs(str);

        Option opt = getOptions.get(COUNT_ARG);
        assert(arg == ((Number) opt.getArg()).longValue());
    }

    public void testLongCountGreaterThan0ArgOption() {
        int arg = 8;
        StringBuffer argBuf = new StringBuffer();
        for ( int i = 0; i < arg; i++ ) {
            argBuf.append("--");
            argBuf.append(COUNT_ARG);
            argBuf.append(" ");
        }

        String[] str = new String[] { argBuf.toString() };
        getOptions.setArgs(str);

        Option opt = getOptions.get(COUNT_ARG);

        assert(arg == ((Number) opt.getArg()).longValue());
    }

    private class OptionMethod extends AbstractOptionMethod {

        public OptionMethod(Option option) {
            super(option);
        }

        public void run() {
            /* will throw NullPointException if getOption returns null */
            try {
                flagMap.put(getOption().getName(), Boolean.TRUE);

            } catch (NullPointerException e) {
                System.err.println("NullPointException thrown");
                System.err.println("getOption()=" + getOption());
                System.err.println("toString()=" + toString());

                throw e;
            }
        }
    }

//    public static Test suite() {
//        return suite(GetoptTest.class);
//    }

    public static void main(String[] args) {
        TestCase.main(new String[] { "jrowe.test.getopt.GetoptTest" });
    }

}

